#include <windows.h>
#include <math.h>
#include <stdint.h>
#include <stdio.h>

#define CONFIGFILE "./TQCam.ini"

#define DEG_TO_RAD (M_PI / 180.0)

typedef struct {
	void *vtable;
	float default_distance;
	float default_pitch;
	float distance;
	float yaw;
	float pitch;
	float fov;
	float far_plane;
	float near_plane;
} WorldCamera;

static const char ret4[3] = { 0xC2, 0x04, 0x00 };
static const char zoom[9] = { 0x55, 0x8B, 0xEC, 0xD9, 0x45, 0x08, 0x56, 0xD8, 0x0D };
static const float TWO_PI = 2.0 * M_PI;

static int key = VK_MBUTTON; 
static float stepx = M_PI / 900.0;
static float stepy = M_PI / 900.0;
static float min_pitch = -89.0 * DEG_TO_RAD;
static float max_pitch = 89.0 * DEG_TO_RAD;
static float stepz = 0.1;

static WorldCamera *camera = 0;

BOOL WINAPI DllMain(HINSTANCE hinst, DWORD reason, LPVOID ptr)
{
	static void *set_yaw = 0;
	static void *set_pitch = 0;
	static void *zoom_ptr = 0;
	static char saved_yaw[3];
	static char saved_pitch[3];
	static char saved_zoom[4];
	char buffer[32];
	HMODULE dll;
	void (_stdcall *set_extents)(float, float, float, float);
	HANDLE proc = GetCurrentProcess();

	switch(reason)
	{
	case DLL_PROCESS_ATTACH:
		set_extents = 0;

		dll = GetModuleHandle("game.dll");
		if (dll)
		{
			void **ptr = (void *)GetProcAddress(dll, "?gGameEngine@GAME@@3PAVGameEngine@1@A");
			if (ptr)
			{
				WorldCamera *(_fastcall *fn)(void *);
				fn = (void *)GetProcAddress(dll, "?GetCamera@GameEngine@GAME@@QAEPAVGameCamera@2@XZ");
				if (fn)
				{
					camera = fn(*ptr);
				}
			}

			set_extents = (void *)GetProcAddress(dll, "?SetMovementExtents@GameCamera@GAME@@QAEXMMMM@Z");

			char *zp = (void *)GetProcAddress(dll, "?Zoom@GameCamera@GAME@@QAEXM@Z");
			if (zp && memcmp(zp, zoom, 9) == 0)
			{
				float *stepp = &stepz;
				zoom_ptr = zp + 9;
				ReadProcessMemory(proc, zoom_ptr, saved_zoom, 4, 0);
				WriteProcessMemory(proc, zoom_ptr, &stepp, 4, 0);
			}
		}

		dll = GetModuleHandle("engine.dll");
		if (dll)
		{
			set_yaw = (void *)GetProcAddress(dll, "?SetCameraYaw@WorldCamera@GAME@@QAEXM@Z");
			if (set_yaw)
			{
				ReadProcessMemory(proc, set_yaw, saved_yaw, 3, 0);
				WriteProcessMemory(proc, set_yaw, ret4, 3, 0);
			}

			set_pitch = (void *)GetProcAddress(dll, "?SetCameraPitch@WorldCamera@GAME@@QAEXM@Z");
			if (set_pitch)
			{
				ReadProcessMemory(proc, set_pitch, saved_pitch, 3, 0);
				WriteProcessMemory(proc, set_pitch, ret4, 3, 0);
			}
		}

		if (camera)
		{
			GetPrivateProfileString("Settings", "MinPitch", "-89", buffer, 32, CONFIGFILE);
			min_pitch = atof(buffer) * DEG_TO_RAD;
			GetPrivateProfileString("Settings", "MaxPitch", "89", buffer, 32, CONFIGFILE);
			max_pitch = atof(buffer) * DEG_TO_RAD;
			GetPrivateProfileString("Settings", "FOV", "90", buffer, 32, CONFIGFILE);
			camera->fov = atof(buffer) * (M_PI / 360.0);
			GetPrivateProfileString("Settings", "FarClip", "150", buffer, 32, CONFIGFILE);
			camera->far_plane = atof(buffer);
			GetPrivateProfileString("Settings", "HSensitivity", "0.2", buffer, 32, CONFIGFILE);
			stepx = atof(buffer) * DEG_TO_RAD;
			GetPrivateProfileString("Settings", "VSensitivity", "0.2", buffer, 32, CONFIGFILE);
			stepy = atof(buffer) * DEG_TO_RAD;
			GetPrivateProfileString("Settings", "Zoom", "10", buffer, 32, CONFIGFILE);
			stepz = 1.0 / atof(buffer);
			key = GetPrivateProfileInt("Settings", "Key", VK_MBUTTON, CONFIGFILE);

			if (set_extents)
			{
				float min, max;

				GetPrivateProfileString("Settings", "MinDistance", "10", buffer, 32, CONFIGFILE);
				min = atof(buffer);
				GetPrivateProfileString("Settings", "MaxDistance", "40", buffer, 32, CONFIGFILE);
				max = atof(buffer);

				asm("movl %0, %%ecx" : : "r"(camera) : "%ecx");
				set_extents(min, max, 30.0, 50.0);
			}
		}
		break;

	case DLL_PROCESS_DETACH:
		if (GetModuleHandle("game.dll"))
		{
			if (zoom_ptr)
			{
				WriteProcessMemory(proc, zoom_ptr, saved_zoom, 4, 0);
			}
		}

		if (GetModuleHandle("engine.dll"))
		{
			if (set_yaw)
			{
				WriteProcessMemory(proc, set_yaw, saved_yaw, 3, 0);
			}
			if (set_pitch)
			{
				WriteProcessMemory(proc, set_pitch, saved_pitch, 3, 0);
			}
		}
		break;
	}

	return TRUE;
}

LRESULT CALLBACK TQC_MouseHook(int code, WPARAM wp, LPARAM lp)
{
	static BOOL btn = 0;
	static LONG x, y;
	PMOUSEHOOKSTRUCT m = (PMOUSEHOOKSTRUCT)lp;

	if (camera && code == HC_ACTION)
	{
		if (GetKeyState(key) < 0)
		{
			if (btn)
			{
				LONG dx = m->pt.x - x;
				LONG dy = m->pt.y - y;

				if (dx)
				{
					float yaw = camera->yaw - dx * stepx;
					while (yaw < 0.0) yaw += TWO_PI;
					while (yaw > TWO_PI) yaw -= TWO_PI;
					camera->yaw = yaw;
				}

				if (dy)
				{
					float pitch = camera->pitch + dy * stepy;
					if (pitch < min_pitch) pitch = min_pitch;
					if (pitch > max_pitch) pitch = max_pitch;
					camera->pitch = pitch;
				}
			}

			btn = 1;
			x = m->pt.x;
			y = m->pt.y;
		}
		else
		{
			btn = 0;
		}
	}

	return CallNextHookEx(0, code, wp, lp);
}
